import os
import sys
import time
import uuid
import math

from functools import partial

import rclpy
from rclpy.action import ActionClient
from rclpy.node import Node

from rclpy.duration import Duration

from threading import RLock
from typing import Any
from rclpy.task import Future

from std_srvs.srv import Trigger
from geographic_msgs.msg import GeoPoint
from geometry_msgs.msg import PoseStamped, Pose
from unique_identifier_msgs.msg import UUID

from lrs_util.coordtrans import CoordTrans
from lrs_util.util import quaternion_from_euler, euler_from_quaternion

from lrs_msgs_common.action import FlyPath, Land, TakeOff

from lrs_exec.command import InAirCheckClient, GetCurrentPositionClient
from lrs_exec.command import FlyPathActionClient, TakeOffActionClient, LandActionClient

from optparse import OptionParser
parser = OptionParser()
parser.add_option ("", "--get-current-position", action="store_true", dest="get_current_position", help="GetCurrentPosition")
parser.add_option ("", "--in-air-check", action="store_true", dest="in_air_check", help="InAirCheck")
parser.add_option ("", "--fly-to", action="store_true", dest="fly_to", help="FlyTo")
parser.add_option ("", "--land", action="store_true", dest="land", help="Land")
parser.add_option ("", "--take-off", action="store_true", dest="take_off", help="Take off")
parser.add_option ("-x", "--pos-x", action="store", type="float", dest="x", default=30.0, help="x")
parser.add_option ("-y", "--pos-y", action="store", type="float", dest="y", default=20.0, help="y")
parser.add_option ("-z", "--pos-z", action="store", type="float", dest="z", default=10.0, help="z")
parser.add_option ("-s", "--speed", action="store", type="float", dest="speed", default=7.0, help="speed")


parser.add_option ("", "--trajectory", action="store_true", dest="trajectory", help="Trajectory")
parser.add_option ("", "--apply-transform", action="store_true", dest="apply_transform", help="Apply Transform")
parser.add_option ("", "--yaw", action="store_true", dest="yaw", help="Yaw")
parser.add_option ("", "--yaw-degree", action="store", type="float", dest="yaw_degree", default=0.0, help="Yaw in degrees")
parser.add_option ("", "--uuid", action="store", type="string", dest="uuid", default="", help="uuid")
parser.add_option ("", "--signal", action="store", type="string", dest="signal", default="enough", help="$enough")
parser.add_option ("", "--signal-goal", action="store_true", dest="signal_goal", help="SignalGoal")
(options, args) = parser.parse_args(rclpy.utilities.remove_ros_args(sys.argv))


def get_pose_stamped(node, x, y, z, yaw, frame_id="world"):
    pose = PoseStamped()
    pose.header.frame_id = frame_id
    pose.header.stamp = node.get_clock().now().to_msg()
    pose.pose.position.x = x
    pose.pose.position.y = y
    pose.pose.position.z = z
    q = quaternion_from_euler(0.0, 0.0, yaw)
    pose.pose.orientation.x =  q[0]
    pose.pose.orientation.y =  q[1]
    pose.pose.orientation.z =  q[2]
    pose.pose.orientation.w =  q[3]
    return pose

def get_pose(node, x, y, z, yaw):
    pose = Pose()
    pose.position.x = x
    pose.position.y = y
    pose.position.z = z
    q = quaternion_from_euler(0.0, 0.0, yaw)
    pose.orientation.x =  q[0]
    pose.orientation.y =  q[1]
    pose.orientation.z =  q[2]
    pose.orientation.w =  q[3]
    return pose

def get_body(node, x, y, z, yaw_in_deg, frame_id="body"):            
    client = ApplyTransformClient(node, shutdown=True)
    pose = get_pose_stamped(node, x, y, z, math.radians(yaw_in_deg))
    client.send_request(frame_id, pose)

def trajectory(node, duration, precise_positioning, pose):
    action_client = TrajectoryActionClient(node, shutdown=True, callback_group=None)
    goal_uuid = action_client.send_goal(pose, duration=duration, precise_positioning=precise_positioning)


    
def main(args=None):
    rclpy.init(args=args)

    coordtrans = CoordTrans(location=os.environ.get("LRS_LOCATION"))

    node = Node("ardu_command")

    if options.fly_to:
        (lon, lat, alt) = coordtrans.world_to_wgs84(options.x, options.y, options.z)
        gp = GeoPoint()
        gp.altitude = alt
        gp.latitude = lat
        gp.longitude = lon
        
        action_client = FlyPathActionClient(node, shutdown=True)
        goal_uuid = action_client.send_goal([gp], options.speed)

    if options.land:
        action_client = LandActionClient(node, shutdown=True)
        action_client.send_goal()

    if options.take_off:
        action_client = TakeOffActionClient(node, shutdown=True)
        action_client.send_goal()

    if options.in_air_check:
        client = InAirCheckClient(node, shutdown=True)
        client.send_request()

    if options.get_current_position:
        client = GetCurrentPositionClient(node, shutdown=True)
        client.send_request()

    rclpy.spin(node)        

if __name__ == '__main__':
    main()
        
