import os
import collections.abc
import collections
collections.MutableMapping = collections.abc.MutableMapping

import math
import time
import rclpy

from rclpy.callback_groups import ReentrantCallbackGroup
from rclpy.duration import Duration
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node

from lrs_msgs_common.msg import GeoPointList, TakeoffInfo
from visualization_msgs.msg import Marker
from std_msgs.msg import ColorRGBA
from geometry_msgs.msg import Point

from tf2_ros import TransformException
from tf2_ros.buffer import Buffer
from tf2_ros.transform_listener import TransformListener
import tf2_geometry_msgs

from functools import partial

from rclpy.action import ActionServer

from lrs_msgs_common.srv import InAirCheck, GetHomePosition, GetCurrentPosition, SignalGoal, GimbalCheck
from lrs_msgs_common.action import FlyPath, Land, TakeOff

from dronekit import connect, VehicleMode, LocationGlobal, Command, mavutil

from lrs_util.coordtrans import CoordTrans

class ArduManager(Node):
    def __init__(self, node_name):
        super().__init__(node_name)
        self.group = ReentrantCallbackGroup()

        self.declare_parameter("host", "localhost")
        self.declare_parameter("port", 14552)
        self.declare_parameter("terra", False)


        self.host = self.get_parameter("host").value
        self.port = self.get_parameter("port").value
        self.terra = self.get_parameter("terra").value
        self.angle_offset = 0.0
        if self.terra:
            self.angle_offset = -90.0

        #self.tf_buffer = Buffer()
        #self.tf_listener = TransformListener(self.tf_buffer, self)

        self.in_air_check_srv = self.create_service(InAirCheck, "in_air_check", self.in_air_check, callback_group=self.group)

        self.get_current_position_srv = self.create_service(GetCurrentPosition, "get_current_position", self.get_current_position, callback_group=self.group)

        self.take_off_action_server = ActionServer(self, TakeOff, "take_off",
                                                   self.take_off_execute_callback)

        self.land_action_server = ActionServer(self, Land, "land",
                                               self.land_execute_callback)

        self.fly_path_server = ActionServer(self, FlyPath, "fly_path",
                                            self.fly_path_execute_callback)

        connection_string = f"tcp:{self.host}:{self.port}"
        print("TRYING TO CONNECT TO:", connection_string)
        self.vehicle = connect(connection_string, wait_ready=True)
        print("CONNECTED")

        self.coordtrans = CoordTrans(location=os.environ.get("LRS_LOCATION"))

        self.takeoff_info_pub = self.create_publisher(TakeoffInfo, "takeoff_info", 10, callback_group=self.group)
        
        self.timer = self.create_timer(1.0, self.timer_callback, callback_group=self.group)


    def timer_callback(self):
        ### self.vehicle.parameters["SIM_BATT_CAPACITY"]=10000
        calt = self.vehicle.location.global_frame.alt
        rcalt = self.vehicle.location.global_relative_frame.alt
        home = self.vehicle.home_location        
        #self.get_logger().info(f'Global frame alt: {calt}')
        #self.get_logger().info(f'Global relative frame alt: {rcalt}')
        #self.get_logger().info(f'Home location: {home}')        
        msg = TakeoffInfo()
        msg.latitude = 0.0
        msg.longitude = 0.0
        msg.height_above_ellipsoid = 29.7
        msg.gps_altitude = 29.7
        self.takeoff_info_pub.publish(msg)
            

    def is_rover(self):
        version = self.vehicle.version
        (name, number) = str(version).split("-")
        if name == "APM:Rover":
            return True
        else:
            return False

    def in_air_check(self, request, response):
        try:
            response.result = False
            response.success = False

            response.result = self.vehicle.armed  # Strange... but seems to work
            
            response.success = True
        except Exception as ex:
            self.get_logger().info(f'Could not transform {to_frame_rel} to {from_frame_rel}: {ex}')
            response.success = False
            response.message = f'Exception: {ex}'
        return response
        

    def get_current_position(self, request, response):
        try:
            response.yaw_in_deg = 0.0
            response.success = False

            clat = self.vehicle.location.global_frame.lat
            clon = self.vehicle.location.global_frame.lon
            calt = self.vehicle.location.global_frame.alt + self.coordtrans.geoid_height
            rcalt = self.vehicle.location.global_relative_frame.alt

            (x, y, z) = self.coordtrans.wgs84_to_world(clon, clat, calt, angle_deg=self.angle_offset)
            response.point.x = x
            response.point.y = y
            response.point.z = z

            response.geopoint.latitude = clat
            response.geopoint.longitude = clon
            response.geopoint.altitude = calt
            
            
            response.success = True
        except Exception as ex:
            self.get_logger().info(f'get_current_position exception: {ex}')
            response.success = False
            response.message = f'Exception: {ex}'
        return response
        

    def take_off_execute_callback(self, goal_handle):
        self.get_logger().info('Executing goal take_off...')
        print("GOAL_HANDLE TAKE OFF:", goal_handle, type(goal_handle))
        
        result = TakeOff.Result()
        

        while not self.vehicle.is_armable:
            self.get_logger().info(f'Waiting for vehicle to initialise...')
            time.sleep(1)
            
        self.get_logger().info(f'Arming motors')            

        # Copter should arm in GUIDED mode
        self.vehicle.mode    = VehicleMode("GUIDED")
        self.vehicle.armed   = True

        # Confirm vehicle armed before attempting to take off
        while not self.vehicle.armed:
            self.get_logger().info(f'Waiting for arming...')
            time.sleep(1)
        
        if self.is_rover():
            result.success = False
            result.message = "A Rover cannot take off"
            return result

        self.get_logger().info(f'Taking off!')
        toh = 3.0
        # toh = self.params.get_value("height-above-takeoff", default=toh)
        target_altitude = toh
        self.vehicle.simple_takeoff(target_altitude) # Take off to target altitude

        feedback_msg = TakeOff.Feedback()
        
        while True:
            self.get_logger().info(f'Altitude: {self.vehicle.location.global_relative_frame.alt}')
            self.pos_feedback(goal_handle, feedback_msg)
            
            #Break and return from function just below target altitude.
            if self.vehicle.location.global_relative_frame.alt >= target_altitude*0.95:
                self.get_logger().info(f'Reached target altitude')
                break
            time.sleep(1)        

        self.get_logger().info(f'finish take-off')

        goal_handle.succeed()
        result.success = True
        result.message = ""
        return result

    def pos_feedback(self, goal_handle, feedback_msg):
        clat = self.vehicle.location.global_frame.lat
        clon = self.vehicle.location.global_frame.lon
        calt = self.vehicle.location.global_frame.alt
        rcalt = self.vehicle.location.global_relative_frame.alt
        home = self.vehicle.home_location        
        self.get_logger().info(f'Global frame alt: {calt}')
        self.get_logger().info(f'Global relative frame alt: {rcalt}')
        self.get_logger().info(f'Home location: {home}')        
        (x, y, z) = self.coordtrans.wgs84_to_world(clon, clat,
                                                   calt+self.coordtrans.get_sea_level_height(), angle_deg=self.angle_offset)
        feedback_msg.current_geopoint.latitude = clat
        feedback_msg.current_geopoint.longitude = clat
        feedback_msg.current_geopoint.altitude = clat
        feedback_msg.current_point.x = x
        feedback_msg.current_point.y = y
        feedback_msg.current_point.z = z
        goal_handle.publish_feedback(feedback_msg)
        
        
    def land_execute_callback(self, goal_handle):
        self.get_logger().info('Executing goal land...')
        print("GOAL_HANDLE LAND:", goal_handle, type(goal_handle))        

        if self.is_rover():
            self.get_logger().info(f'Disarm')
            self.vehicle.armed = False
        else:
            self.get_logger().info(f'Land')                        
            self.vehicle.mode = VehicleMode("LAND")

        feedback_msg = Land.Feedback()
        
        while self.vehicle.armed:
            self.get_logger().info(f'Altitude: {self.vehicle.location.global_relative_frame.alt}')

            self.pos_feedback(goal_handle, feedback_msg)
            time.sleep(1)        

        goal_handle.succeed()
        result = Land.Result()
        result.success = True
        result.message = ""
        return result

    def is_armed(self):
        return self.vehicle.armed
    
    def is_manual(self):
        return self.is_armed() and self.vehicle.mode.name in [
            "ACRO",
            "ALT_HOLD",
            "AUTOTUNE",
            "AUTO_TUNE",
            "DRIFT",
            "FBWA",
            "FBWB",
            "FLIP",
            "MANUAL",
            "POSHOLD",
            "STABLILIZE",
            "SPORT",
            "SIMPLE",
            "STEERING",
            "THROTTLE",
            "TRAINING",
        ]

    def is_auto(self) -> bool:
        return self.is_armed() and self.get_mode_name() in [
            "AUTO",
            "CIRCLE",
            "CRUISE",
            "DOCK",
            "FOLLOW",
            "INITIALIZE",
            "LAND",
            "RTL",
            "SMARTRTL",
            "SMART_RTL",
            "SURFACE",
            "TAKEOFF",
            "TERRAIN",
            "QRTL",
        ]

    def start_mission(self):
        if self.is_manual():
            print("IN MANUAL MODE, CANNOT FLY")
            return False
        self.vehicle.commands.upload()
        self.vehicle.commands.wait_ready()
        print("MISSION UPLOADED")
        res = self.set_mode("AUTO")
        return res

    def set_mode(self, new_mode: str) -> bool:
        print(f"{self.vehicle.mode.name} {new_mode}")
        if self.vehicle.mode.name == new_mode:
            return True
        else:
            loop = 0
            timeout_count = 5

            while self.vehicle.mode.name != new_mode:
                self.vehicle.mode = VehicleMode(new_mode)
                loop += 1
                if loop > timeout_count:
                    break
                time.sleep(1)
        return self.vehicle.mode.name == new_mode
    
    def pause_mission(self):
        pause_mode = None
        if self.is_rotary():
            paus_mode = "BRAKE"
        elif self.is_fixed_wing():
            paus_mode = "LOITER"
        elif self.is_subsurface():
            paus_mode = "POSHOLD"
        elif self.is_ground():
            paus_mode = "LOITER"
        else:
            paus_mode = "LOITER"
        if self.set_mode(paus_mode):
            return True
        return False
        
    def continue_mission(self):
        res = False
        if self.is_subsurface():
            res = self.set_mode("GUIDED")
        else:
            res = self.set_mode("AUTO")
        return res

    def stop_mission(self):
        self.clear_mission()
        return self.pause_mission()
        
    def clear_mission(self):
        self.vehicle.commands.download()
        self.vehicle.commands.wait_ready()
        self.vehicle.commands.clear()
        res = self.set_mode("LOITER")        
        self.vehicle.commands.upload()
        self.vehicle.commands.wait_ready()
        print("MISSION CLEARED")
    
    def fly_path_execute_callback(self, goal_handle):
        self.get_logger().info('Executing fly path...')
        print("SPEED", goal_handle.request.speed)
        print("GOALHANDLE GEOPOINTS", goal_handle.request.geopoints)

        gps = goal_handle.request.geopoints
        speed = goal_handle.request.speed

        if self.is_rover():
            self.vehicle.groundspeed = speed
        else:
            self.vehicle.airspeed = speed

        #gp = gps[0]
        #print("FLYTO:", alt)        
        #point = LocationGlobal(lat, lon, alt)
        #self.vehicle.simple_goto(point)
        
        home = self.vehicle.home_location
        
        self.clear_mission()

        for gp in gps:
            lat = gp.latitude
            lon = gp.longitude
            alt = gp.altitude - self.coordtrans.get_sea_level_height()
            (x0, y0, z0) = self.coordtrans.wgs84_to_world(lon, lat, gp.altitude, angle_deg=self.angle_offset)
            ###z0 = alt

            wp_cmd: Command = Command(
                0,  # target system
                0,  # target component
                0,  # sequence
                mavutil.mavlink.MAV_FRAME_GLOBAL,  # frame
                mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,  # command
                0,  # current
                0,  # autucontinue
                0,  # param1: Speed type (0=Airspeed, 1=Ground Speed)
                0,  # param2: Speed in m/s
                0,  # param3: Throttle (ignored)
                0,  # param4: Absolute or relative (ignored)
                lat,  # 11: latitude in degrees
                lon,  # 12: longitude in degrees
                alt,  # 13: altitude in meters
            )
            self.vehicle.commands.add(wp_cmd)
        
        res = self.start_mission()

        if not res:
            goal_handle.succeed()            
            result = FlyPath.Result()
            result.success = False
            result.message = "Start mission failed"
            result.abort_flag = False
            result.enough_flag = False
            return result
            
        
        feedback_msg = FlyPath.Feedback()

        d2 = 10.0
        while d2 > 0.4:
            time.sleep(1)
            clat = self.vehicle.location.global_frame.lat
            clon = self.vehicle.location.global_frame.lon
            calt = self.vehicle.location.global_frame.alt
            rcalt = self.vehicle.location.global_relative_frame.alt
            self.get_logger().info(f'Global frame alt: {calt}')
            self.get_logger().info(f'Global relative frame alt: {rcalt}')
            self.get_logger().info(f'Home location: {home}')
            (x, y, z) = self.coordtrans.wgs84_to_world(clon, clat, calt+self.coordtrans.get_sea_level_height(), angle_deg=self.angle_offset)
            self.get_logger().info(f'WORLD POSITION: {x} {y} {z}')
            dx = x-x0
            dy = y-y0
            dz = z-z0
            if self.is_rover():
                d2 = dx*dx + dy*dy
                self.get_logger().info(f'Distance Rover: {dx} {dy} - {d2}')
            else:
                d2 = dx*dx + dy*dy + dz*dz
                self.get_logger().info(f'Distance Copter: {dx} {dy} {dz} - {d2}')

            self.pos_feedback(goal_handle, feedback_msg)
            
        goal_handle.succeed()            
        result = FlyPath.Result()
        result.success = True
        result.message = ""
        result.abort_flag = False
        result.enough_flag = False
        return result
        
    

def main(args=None):
    rclpy.init(args=args)

    print("main1")
    executor = MultiThreadedExecutor(num_threads=8)
    
    ctm = ArduManager("ardu_manager")

    executor.add_node(ctm)
    print("main2")
    
    try:
        print("Spinning ardu_manager")
        executor.spin()
        print("main3")
    
    except KeyboardInterrupt:
        pass

    ctm.destroy_node()
    executor.shutdown()    
    rclpy.shutdown()
    
        
if __name__ == '__main__':
    main()
