import collections.abc
import collections
collections.MutableMapping = collections.abc.MutableMapping

import os
import time
import sys
import math
import threading
import rclpy
from rclpy.node import Node
from rclpy.executors import MultiThreadedExecutor
from rclpy.callback_groups import ReentrantCallbackGroup

from dronekit import connect, VehicleMode

from geographic_msgs.msg import GeoPoint
from geometry_msgs.msg import PoseStamped
from std_msgs.msg import Float32
from lrs_util.coordtrans import CoordTrans
from lrs_util.locations import Locations
from lrs_util.util import quaternion_from_euler

class ArduPilotConnect(Node):

    def __init__(self, vehicle):
        super().__init__('turtle')
        self.group = ReentrantCallbackGroup()
        self.vehicle = vehicle
        self.coordtrans = CoordTrans(location=os.environ.get("LRS_LOCATION"))
        (lon, lat, alt) = self.coordtrans.world_to_wgs84(0, 0, 0.0)
        self.wgs84_origin_alt = alt
        
        self.position_publisher = self.create_publisher(GeoPoint, 'sensor/position', 1)
        self.heading_publisher = self.create_publisher(Float32, 'sensor/heading', 1)
        self.course_publisher = self.create_publisher(Float32, 'sensor/course', 1)
        self.speed_publisher = self.create_publisher(Float32, 'sensor/speed', 1)

        timer_period = 1/50.0  # seconds
        self.timer = self.create_timer(timer_period, self.timer_callback)
        self.i = 0
        self.heading = 90.0
        self.course = 50.0
        self.speed = 7.0
        self.location = "granso"
        self.coordtrans = CoordTrans(location=self.location)        
        self.x = 0.0
        self.y = 0.0
        self.z = 0.0
        self.speed = 0.0
        self.course = 0.0
        self.yaw = 0.0
        self.linear_velocity = 0.0
        self.angular_velocity = 0.0
        self.have_pose = False
        self.goal_x = 0.0
        self.goal_y = 0.0
        self.n_odom_calls = 0
        self.lat = 0.0
        self.lon = 0.0
        self.alt = 0.0
        self.pose = None

        #self.odom_sub = self.create_subscription(Odometry, "odom", self.odom_callback, 10, callback_group=self.group)
        self.pose_pub = self.create_publisher(PoseStamped, "pose", 1)

    def sensor_level(self):
        try:
            fix = self.vehicle.gps_0.fix_type
            num_sat = self.vehicle.gps_0.satellites_visible
            yaw = self.vehicle.attitude.yaw
            vx = self.vehicle.velocity[0]
            vy = self.vehicle.velocity[1]
            vz = self.vehicle.velocity[2]
            speed = math.sqrt(vx*vx+vy*vy+vz*vz)
            heading = math.degrees(yaw)
            course = heading
            if speed > 0.1:
                course = math.degrees(math.atan2(vy, vx))

            hdop = self.vehicle.gps_0.eph
            hdop = self.vehicle.gps_0.epv
                
            self.lat = self.vehicle.location.global_frame.lat
            self.lon = self.vehicle.location.global_frame.lon
            msl_alt = self.vehicle.location.global_frame.alt

            
            if True:
                print(f'GPS_INFO: {self.vehicle.gps_0} - {type(self.vehicle.gps_0)}')
                print(" Battery: %s" % self.vehicle.battery)
                print(" Last Heartbeat: %s" % self.vehicle.last_heartbeat)
                print(" Is Armable?: %s" % self.vehicle.is_armable)
                print(f'Armed: {self.vehicle.armed}')
                print(" System status: %s" % self.vehicle.system_status.state)
                print(" Mode: %s" % self.vehicle.mode.name)    # settable

                print(f'Global position (global): {self.vehicle.location.global_frame}')
                print(f'Global position (global_local): {self.vehicle.location.global_relative_frame}')
                print(f'Global position (local): {self.vehicle.location.local_frame}')
                print(f'Heading (north=0): {self.vehicle.heading}')
                print(f'Course: {course}')                    
                print(f'Attitude: {self.vehicle.attitude}')
                print(f'Velocity: {self.vehicle.velocity}')
                print(f'Yaw Degrees: {math.degrees(yaw)}')
                print(f'Speed: {speed}')
                print(f'Releasetype: {self.vehicle.version.release_type()}')
                print(f'Terrain: {self.vehicle.capabilities.terrain}')
                print(f'Capabilities: {self.vehicle.capabilities.set_attitude_target}')
                print(f'FrameClass: {self.vehicle.parameters["FRAME_CLASS"]}')
                print(f'Version: {self.vehicle.version.vehicle_type} - {self.vehicle.version.autopilot_type} - {str(self.vehicle.version)}')
                #for key, value in self.vehicle.parameters.iteritems():
                #    print(f' Key: {key} Value: {value}')



            
            # print("Publish Sensor Topics")
            
            if num_sat > 5 and fix > 2:
                heading_msg = Float32()
                heading_msg.data = heading
                self.heading_publisher.publish(heading_msg)
                
                course_msg = Float32()
                course_msg.data = course
                self.course_publisher.publish(course_msg)

                speed_msg = Float32()
                speed_msg.data = speed
                self.speed_publisher.publish(speed_msg)

                gp_msg = GeoPoint()
                gp_msg.latitude = self.lat
                gp_msg.longitude = self.lon
                gp_msg.altitude = msl_alt + self.wgs84_origin_alt
                self.position_publisher.publish(gp_msg)

                (x, y, z) = self.coordtrans.wgs84_to_world(self.lon, self.lat, self.alt)
                self.pose = PoseStamped()
                
                self.pose.header.stamp = self.get_clock().now().to_msg()
                self.pose.header.frame_id = "world"
                self.pose.pose.position.x = x
                self.pose.pose.position.y = y
                self.pose.pose.position.z = z
                att = self.vehicle.attitude
                q = quaternion_from_euler(att.roll, att.pitch, att.yaw)
                self.pose.pose.orientation.x = q[0]
                self.pose.pose.orientation.y = q[1]
                self.pose.pose.orientation.z = q[2]
                self.pose.pose.orientation.w = q[3]

                
        except Exception as exc:
            self.get_logger().error(f'EXCEPTION sensor_level: {type(exc)} - {exc}')            
                
    def timer_callback(self):
        self.sensor_level()
        if self.pose:
            self.pose_pub.publish(self.pose)

def main(args=None):
    print('Hi from tbot_connect.')

    rclpy.init(args=args)

    executor = MultiThreadedExecutor(num_threads=8)

    connection_string = "tcp:localhost:14552"

    print(f'Connecting to vehicle on: {connection_string}')
    vehicle = connect(connection_string, wait_ready=True)
    
    
    con = ArduPilotConnect(vehicle)
    executor.add_node(con)


    try:
        executor.spin()
    except KeyboardInterrupt:
        pass

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    executor.shutdown()
    rclpy.shutdown()
    
if __name__ == '__main__':
    main()
    
        
