import os
import time
import rclpy

from lrs_exec.ticked_executor import TickedExecutor
from lrs_exec.ticked_executor_factory import TickedExecutorFactory

from lrs_exec.command import FlyPathActionClient, GetCurrentPositionClient

from geographic_msgs.msg import GeoPoint
from lrs_util.coordtrans import CoordTrans

from lrs_ardupilot.executors.mission_executor import MissionExecutor

class FlyToExecutor(MissionExecutor):
    def __init__(self, node, id):
        try:
            super().__init__(node, id)
            self.add_resource_to_lock("fly")
            self.action_client = FlyPathActionClient(node, callback_group=node.group)
            # self.get_home_position_client = GetHomePositionClient(node)
            self.get_current_position_client = GetCurrentPositionClient(node)
            # self.signal_goal_client = SignalGoalClient(node)
            # self.planner = Planner(node)
            self.coordtrans = CoordTrans(location=os.environ.get("LRS_LOCATION"))
            self.node.get_logger().info(f'{__name__} created XXX')
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION fly_to: {type(exc)} - {exc}')

    def get_commanded_speed(self):
        try:
            speed = self.params.get_value("speed", default="standard")
            print("SPEED:", speed)
            self.speed = speed
            commanded_speed = 7.0
            if speed == "standard":
                commanded_speed = 7.0
            if speed == "fast":
                commanded_speed = 15.0
            if speed == "slow":
                commanded_speed = 3.0
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION fly_to: {type(exc)} - {exc}')
        return commanded_speed
    
    def do_work_begin(self):
        try:
            self.node.get_logger().info(f'{__name__} do_work_begin YY')

            self.fly_when_no_plan_found = True
            
            gp = self.params.get_value("p")
            self.node.get_logger().info(f'GP: {gp}')

            commanded_speed = self.get_commanded_speed()
            self.node.get_logger().info(f'COMMANDED SPEED: {commanded_speed}')

            climb_flag = False

            fly_home_flag = self.params.get_value("fly_home_flag", default=False)
            self.node.get_logger().info(f'fly_home_flag: {fly_home_flag}')

            fly_home_height = self.params.get_value("fly_home_height", default=30.0)
            self.node.get_logger().info(f'fly_home_height: {fly_home_height}')

            self.get_current_position_client.send_request()
            while not self.get_current_position_client.result:
                time.sleep(0.1)
            self.node.get_logger().info(f'current GP: {self.get_current_position_client.result}')
            cgp = self.get_current_position_client.result.geopoint

            if climb_flag:
                gp.latitude = cgp.latitude
                gp.longitude = cgp.longitude
                lon, lat, alt = self.coordtrans.world_to_wgs84(0, 0, fly_home_height)            
                gp.altitude = alt
            

            if not gp:
                self.node.get_logger().error(f'{__name__} fail: do not have position')
                self.fail("Do not have position")
                self.do_work_finished()                
                return

            safety_margin = -1.0
            timeout = 3.0

            if self.same_position(cgp, gp):
                self.node.get_logger().info(f'GOAL POSITION NEAR CURRENT POSITION. Do Nothing!')
                self.do_work_finished()                
                return

            plan_path = []
            #if self.use_motion_planner:
            #    plan_path = self.planner.plan_segment(cgp, gp, timeout, safety_margin)

            if not plan_path:
                if self.fly_when_no_plan_found:
                    self.node.get_logger().info(f'FAILED TO GET PLAN: Flying anyway')
                    plan_path = [cgp, gp]
                else:
                    self.node.get_logger().error(f'{__name__} fail: Could not find a plan')
                    self.fail("Could not find a plan")
                    self.do_work_finished()                    
                    return

            for el in plan_path:
                self.node.get_logger().info(f'PATH: {el}')
            self.display_path(plan_path)
            self.pub_waypoints(plan_path)
            
            self.goal_uuid = self.action_client.send_goal(plan_path[1:], commanded_speed) # Seems like plan_path contains current position, action does not need this position...

            counter = 0
            while not self.action_client.have_goal_response and counter < 100:
                counter += 1
                time.sleep(0.1)

            if not self.action_client.have_goal_response:
                self.fail(f'Manager execution failed in do_work_begin: No goal response in 3 seconds')
                self.do_work_finished()
                return

            if not self.action_client.goal_accepted:
                self.fail(f'Manager execution failed in do_work_begin: Goal not accepted')
                self.do_work_finished()
                return
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION fly_to: {type(exc)} - {exc}')
                

    def do_work(self):
        try:
            if self.action_client.result:
                # check result for failure...
                self.node.get_logger().info(f'fly_path_result: {self.action_client.result}')
                if not self.action_client.result.success:
                    if not self.action_client.result.enough_flag:
                        if self.action_client.result.abort_flag:
                            self.abort_fail(f'Manager execution failed: {self.action_client.result.message}')
                        else:
                            self.fail(f'Manager execution failed: {self.action_client.result.message}')                            
                self.do_work_finished()
            self.handle_signals()
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work: {type(exc)} - {exc}')            
            self.fail(f'Exception: {exc}')
            
class FlyToExecutorFactory(TickedExecutorFactory):
    def __init__(self):
        super().__init__("fly-to")

    def get_executor(self, node, id):
        return self.add_executor(FlyToExecutor(node, id))
        
