import os
import time
import math
import rclpy

from lrs_exec.ticked_executor import TickedExecutor
from lrs_exec.ticked_executor_factory import TickedExecutorFactory

from lrs_exec.command import FlyPathActionClient, TakeOffActionClient, GetCurrentPositionClient

from geographic_msgs.msg import GeoPoint

from lrs_ardupilot.executors.mission_executor import MissionExecutor

class TakeoffExecutor(MissionExecutor):
    def __init__(self, node, id):
        super().__init__(node, id)
        self.action_client = TakeOffActionClient(node, callback_group=node.group)
        self.get_current_position_client = GetCurrentPositionClient(node)
        self.add_resource_to_lock("fly")
        self.node.get_logger().info(f'{__name__} created')

    def do_work_begin(self):
        try:
            self.node.get_logger().info(f'{__name__} do_work_begin')

            self.action_client.send_goal()

            counter = 0
            while not self.action_client.have_goal_response and counter < 100:
                counter += 1
                time.sleep(0.1)

            if not self.action_client.have_goal_response:
                self.fail(f'Manager execution failed in do_work_begin: No goal response in 3 seconds')
                self.do_work_finished()
                return

            if not self.action_client.goal_accepted:
                self.fail(f'Manager execution failed in do_work_begin: Goal not accepted')
                self.do_work_finished()
                return
                
            while not self.action_client.result:
                time.sleep(0.1)

            if not self.action_client.result.success:                
                self.fail(f'Manager execution failed: {self.action_client.result.message}')
                self.do_work_finished()
                return
            
            # climb

            toh = 15.0
            toh = self.params.get_value("height-above-takeoff", default=toh)

            # get current gp

            self.get_current_position_client.send_request()
            while not self.get_current_position_client.result:
                time.sleep(0.1)

            cgp = self.get_current_position_client.result.geopoint
            cp = self.get_current_position_client.result.point
            current_yaw_in_deg = self.get_current_position_client.result.yaw_in_deg
            
            self.node.get_logger().info(f'CURRENT YAW IN DEG: {current_yaw_in_deg}')
            
            # add toh to it

            gp1 = GeoPoint()
            gp1.latitude = cgp.latitude
            gp1.longitude = cgp.longitude
            gp1.altitude = cgp.altitude + toh


            x0, y0, z0  = self.coordtrans.wgs84_to_world(cgp.longitude, cgp.latitude, cgp.altitude)
            x1, y1, z1  = self.coordtrans.wgs84_to_world(gp1.longitude, gp1.latitude, gp1.altitude)
            self.node.get_logger().info(f'CURRENT POINT 0: {x0} {y0} {z0}')
            self.node.get_logger().info(f'CURRENT POINT 1: {x1} {y1} {z1}')

            dx = math.cos(math.radians(current_yaw_in_deg))
            dy = math.sin(math.radians(current_yaw_in_deg))

            x0 += dx*1.0
            y0 += dy*1.0
            x1 += dx*2.0
            y1 += dy*2.0

            # do fly_plan_path
            
            self.node.get_logger().info(f'NEW POINT FORWARD 1 m: {x0} {y0} {z0}')            
            self.node.get_logger().info(f'NEW POINT FORWARD 2 m: {x1} {y1} {z1}')            

            #lon0, lat0, alt0 = self.coordtrans.world_to_wgs84(x0, y0, z0)
            #gp0.latitude = lat0
            #gp0.longitude = lon0
            
            lon1, lat1, alt1 = self.coordtrans.world_to_wgs84(x1, y1, z1)
            gp1.latitude = lat1
            gp1.longitude = lon1

            self.action_client = FlyPathActionClient(self.node, callback_group=self.node.group)
            commanded_speed = 2.0
            self.goal_uuid = self.action_client.send_goal([gp1], commanded_speed)
            
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION start: {type(exc)} - {exc}')

    def do_work(self):
        try:
            if self.action_client.result:
                # check result for failure...
                self.node.get_logger().info(f'fly_path_result: {self.action_client.result}')
                if not self.action_client.result.success:
                    if not self.action_client.result.enough_flag:
                        if self.action_client.result.abort_flag:
                            self.abort_fail(f'Manager execution failed: {self.action_client.result.message}')
                        else:
                            self.fail(f'Manager execution failed: {self.action_client.result.message}')                            
                self.do_work_finished()
            self.handle_signals()
        except Exception as exc:
            self.node.get_logger().error(f'EXCEPTION do_work: {type(exc)} - {exc}')            
            self.fail(f'Exception: {exc}')
        
            
class TakeoffExecutorFactory(TickedExecutorFactory):
    def __init__(self):
        super().__init__("take-off")

    def get_executor(self, node, id):
        return self.add_executor(TakeoffExecutor(node, id))
        
